/*
 * G-code meta compiler
 *
 * Copyright (C) 2013  B. Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "utils.h"
#include "variable.h"

#define NSORTVARS	16	/* More than these nr. of vars should use qsort/bsearch */

typedef struct __scope_t {
	variable_t **vars;
	int nvars;
	int navars;
} scope_t;

scope_t globalscope;

static scope_t *scopes;
static int nscopes;
static int nascopes;

#define ls_vars		(scopes[nscopes-1].vars)
#define ls_nvars	(scopes[nscopes-1].nvars)
#define ls_navars	(scopes[nscopes-1].navars)

#define gs_vars		(globalscope.vars)
#define gs_nvars	(globalscope.nvars)
#define gs_navars	(globalscope.navars)

void printvar(const char *pfx, variable_t *v)
{
	printf("%s: %s: ", pfx, v->id);
	switch(v->v->type) {
	case VAL_UNDEF:		printf("undef"); break;
	case VAL_INT:		printf("int(%d)", v->v->i); break;
	case VAL_FLOAT:		printf("float(%f)", v->v->d); break;
	case VAL_VECTOR:	printf("vector[%d]", v->v->v.n); break;
	case VAL_VECTORLIST:	printf("vectorlist{%d}", v->v->vl.n); break;
	default:		printf("<invalid variable value>"); break;
	}
	printf("\n");
}

void variable_purge(void)
{
	int i;
//printf("variable_purge\n");
	while(nscopes)
		variable_scope_pop();

	if(scopes)
		free(scopes);

	for(i = 0; i < gs_nvars; i++) {
		value_delete(gs_vars[i]->v);
		free(gs_vars[i]);
	}
	free(gs_vars);
}

static int var_cmp(const void *a, const void *b)
{
	return strcmp((*(variable_t **)a)->id, (*(variable_t **)b)->id);
}

const variable_t *variable_lookup(const char *id, int onlyglobal)
{
	int i;
	if(!onlyglobal && nscopes) {
		if(ls_nvars >= NSORTVARS) {
			variable_t var;
			variable_t *vp = &var;
			variable_t **vpp;
			var.id = id;
			vpp = bsearch(&vp, ls_vars, ls_nvars, sizeof(*ls_vars), var_cmp);
			if(vpp)
				return *vpp;
		} else {
			for(i = 0; i < ls_nvars; i++) {
				if(!ls_vars[i])
					continue;
				if(!strcmp(ls_vars[i]->id, id))
					return ls_vars[i];
			}
		}
	}

	if(gs_nvars >= NSORTVARS) {
		variable_t var;
		variable_t *vp = &var;
		variable_t **vpp;
		var.id = id;
		vpp = bsearch(&vp, gs_vars, gs_nvars, sizeof(*gs_vars), var_cmp);
		if(vpp)
			return *vpp;
	} else {
		for(i = 0; i < gs_nvars; i++) {
			if(!gs_vars[i])
				continue;
			if(!strcmp(gs_vars[i]->id, id))
				return gs_vars[i];
		}
	}
	return NULL;
}

void variable_scope_push(void)
{
//printf("variable_scope_push\n");
	testalloc((void **)&scopes, nscopes, &nascopes, sizeof(scopes[0]));
	scopes[nscopes].vars = NULL;
	scopes[nscopes].nvars = 0;
	scopes[nscopes].navars = 0;
	nscopes++;
	testalloc((void **)&ls_vars, ls_nvars, &ls_navars, sizeof(ls_vars[0]));
}

void variable_scope_pop(void)
{
	int i;
	for(i = 0; i < ls_nvars; i++) {
		value_delete(ls_vars[i]->v);
		free(ls_vars[i]);
	}
	free(ls_vars);
	nscopes--;
//printf("variable_scope_pop\n");
}

static variable_t *variable_addref(variable_t *v)
{
	if(nscopes) {
		testalloc((void **)&ls_vars, ls_nvars, &ls_navars, sizeof(ls_vars[0]));
		ls_vars[ls_nvars] = v;
		ls_nvars++;
		if(ls_nvars >= NSORTVARS)
			qsort(ls_vars, ls_nvars, sizeof(*ls_vars), var_cmp);
	} else {
		testalloc((void **)&gs_vars, gs_nvars, &gs_navars, sizeof(gs_vars[0]));
		gs_vars[gs_nvars] = v;
		gs_nvars++;
		if(gs_nvars >= NSORTVARS)
			qsort(gs_vars, gs_nvars, sizeof(*gs_vars), var_cmp);
	}
	return v;
}

const variable_t *variable_new(const char *id, value_t *v)
{
	variable_t *var = calloc(1, sizeof(*var));
	assert(var != NULL);
	assert(id !=  NULL);
	assert(v !=  NULL);
	var->id = id;
	var->v = v;
	variable_addref(var);
//printvar("variable_new", var);
	return var;
}

const variable_t *variable_set(const char *id, value_t *v)
{
	variable_t *var;
	if(!(var = (variable_t *)variable_lookup(id, 0)))
		return variable_new(id, v);
	else {
		value_delete(var->v);
		var->v = v;
	}
	return var;
}

const variable_t *variable_set_local(const char *id, value_t *v)
{
	int i;
	if(!nscopes)
		rtinternal(NULL, "Locally scoped variable '%s' setting without local scope", id);

	for(i = 0; i < ls_nvars; i++) {
		if(!ls_vars[i])
			continue;
		if(!strcmp(ls_vars[i]->id, id))
			break;
	}

	if(i >= ls_nvars)
		return variable_new(id, v);
	else {
		rtwarning(NULL, "Local scope variable '%s' already available in local scope", id);
		return ls_vars[i];
	}
}

